r"""
Plotting functions

.. autosummary::
   :toctree: generated/

   compare_groups_of_contacts
   plot_matrix
   plot_unified_freq_dicts
   plot_unified_distro_dicts
   compare_violins
   color_dict_guesser

"""

from .plots import *