r"""

Container for other sub-modules with lower-level functions.
They are exposed, tested, and documented, so feel free to
use them in your workflow as needed.

.. currentmodule:: mdciao.utils

.. autosummary::
   :nosignatures:
   :toctree: generated/

   residue_and_atom
   bonds
   str_and_dict
   COM
   lists
   contact_matrix
   sequence

"""

from . import str_and_dict
from . import bonds
from . import residue_and_atom
from . import COM
from . import lists
from . import contact_matrix
from . import sequence