r"""

RCSB-PDB web lookups

Relevant links:
 * https://data.rcsb.org/index.html#data-api
 * https://data.rcsb.org/data-attributes.html

Methods not documented here are either private or will be soon,
please note the API is not yet stable.

.. currentmodule:: mdciao.pdb

.. autosummary::
   :toctree: generated/

    pdb2ref
    pdb2traj


"""
from .pdb import *