r"""
Programatic access to the CLI methods

.. currentmodule:: mdciao.cli

.. autosummary::
   :toctree: generated/

   residue_neighborhoods
   interface
   sites
   compare
   residue_selection
   pdb
   fragment_overview
"""
from .cli import *
